/*
 * Copyright (C) 2005 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License.
 * You must obey the GNU General Public License in all respects for all of the code used
 * other than the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.config;

import multex.Exc;
import multex.Failure;
import org.apache.commons.collections.ExtendedProperties;

import javax.servlet.ServletContext;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * @author idefix
 */
public class MirPropertiesConfiguration extends ExtendedProperties {
  private static MirPropertiesConfiguration instance;
  private static ServletContext context;

  private File home;

  private MirPropertiesConfiguration(ServletContext aContext) throws IOException {
    //loading the defaults-config
    super(aContext.getRealPath("/WEB-INF/") + "/default.properties");

    //loading the user-config
    ExtendedProperties userConfig =
      new ExtendedProperties(aContext.getRealPath("/WEB-INF/etc/") + "/config.properties");

    //merging them to one config while overriding the defaults
    this.combine(userConfig);

    home = new File(aContext.getRealPath("/WEB-INF/")+"/");
  }

  public static synchronized MirPropertiesConfiguration instance() {
    if (instance == null) {
      if (context == null) {
        throw new Error("Context was not set");
      }

      try {
        instance = new MirPropertiesConfiguration(context);
      }
      catch (IOException e) {
        throw new Error("cannot load configuration: " + e.toString());
      }
    }

    return instance;
  }

  /**
   * Sets the context.
   *
   * @param context The context to set
   */
  public static void setContext(ServletContext context) {
    MirPropertiesConfiguration.context = context;
  }

  /**
   * Returns the context.
   *
   * @return ServletContext
   */
  public static ServletContext getContext() {
    return context;
  }

  /**
   * Returns all properties in a Map
   *
   * @return Map
   */
  public Map allSettings() {
    Iterator iterator = this.getKeys();
    Map returnMap = new HashMap();
    while (iterator.hasNext()) {
      String key = (String) iterator.next();
      Object o = this.getProperty(key);

      returnMap.put(key, o);
    }

    // ML: hack for now
    if (!returnMap.containsKey("Producer.DocRoot")) {
      returnMap.put("Producer.DocRoot", "");
    }

    return returnMap;
  }

  /**
   * Return mir's home directory.
   * Normally this is the <code>WEB-INF</code> dir of the
   * deployed mir servlets.
   */
  public File getHome() {
    return home;
  }

  /**
   * Returns a file based on a configuration setting.
   *
   * The file may be configured with an absolute path, or
   * it may be a relative path.
   *
   * Relative paths work relative to {@link #home} :
   * normally this is the <code>WEB-INF</code> dir in a
   * deployed java servlet.
   */
  public File getFile(String aKey) {
    String path = getString(aKey);

    File result = new File(path);
    if (result.isAbsolute()) {
      return result;
    }
		return new File(home, path);
  }

  /**
   * @return the vlaue of this property as String
   * @param key the key of this property
   */
  public String getString(String key) {
    return getString(key, "");
  }


  /**
   * @return the value of this property as String
   * @param aKey the key of the property
   * @param aDefaultValue the default value of this property if it is null
   * @see org.apache.commons.collections.ExtendedProperties#getString(java.lang.String, java.lang.String)
   */
  public String getString(String aKey, String aDefaultValue) {
    if (aDefaultValue == null) {
      aDefaultValue = "";
    }
    Object result = getProperty(aKey);

    if (result == null){
      return aDefaultValue;
    }

    if (result instanceof String) {
      return (String) result;
    }

    return result.toString();
  }

  public boolean getBoolean(String aKey, boolean aDefaultValue) {
    try {
      return getBoolean(aKey);
    }
    catch (Throwable t) {
      return aDefaultValue;
    }
  }

  public boolean getBoolean(String aKey) {
    String value = getString(aKey).trim();

    return "1".equals(value) || "y".equalsIgnoreCase(value) ||
           "yes".equalsIgnoreCase(value) || "true".equalsIgnoreCase(value);
  }

  /**
   * Returns a property according to the given key
   * @param key the key of the property
   * @return the value of the property as Object, if no property available it returns a empty String
   */
  public Object getProperty(String key) {
    if (super.getProperty(key) == null) {
      return "";
    }

    return super.getProperty(key);
  }

  /**
   * @author idefix
   */
  public static class PropertiesConfigExc extends Exc {
    /**
     * Constructor for PropertiesConfigExc.
     */
    public PropertiesConfigExc(String msg) {
      super(msg);
    }
  }

  /**
   * @author idefix
   */
  public static class PropertiesConfigFailure extends Failure {
    /**
     * Constructor for PropertiesConfigExc.
     */
    public PropertiesConfigFailure(String msg, Throwable cause) {
      super(msg, cause);
    }
  }
}
