/*
 * Copyright (C) 2005 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License.
 * You must obey the GNU General Public License in all respects for all of the code used
 * other than the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.generator;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Collection;

import mir.log.LoggerWrapper;
import mir.util.GeneratorFormatAdapters;
import mir.util.RewindableIterator;

import org.apache.commons.beanutils.MethodUtils;
import org.apache.commons.beanutils.PropertyUtils;

import freemarker.template.FileTemplateCache;
import freemarker.template.SimpleScalar;
import freemarker.template.Template;
import freemarker.template.TemplateHashModel;
import freemarker.template.TemplateListModel;
import freemarker.template.TemplateMethodModel;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateModelException;
import freemarker.template.TemplateModelRoot;
import freemarker.template.TemplateScalarModel;


public class FreemarkerGenerator implements Generator {
  private Template template;
  private Interceptor interceptor;

  public FreemarkerGenerator(Template aTemplate, Interceptor anInterceptor) {
    template = aTemplate;
    interceptor = anInterceptor;
  }

  public void generate(Object anOutputWriter, Map aValues, LoggerWrapper aLogger) throws GeneratorExc, GeneratorFailure {
    if (!(anOutputWriter instanceof PrintWriter)) {
      throw new GeneratorExc("Writer for a FreemarkerGenerator must be a PrintWriter");
    }

    try {
      template.process((TemplateModelRoot) makeMapAdapter(aValues), (PrintWriter) anOutputWriter);
    }
    catch (Throwable t) {
      aLogger.error("Exception occurred: "+t.getMessage(), t);

      throw new GeneratorFailure(t);
    }
  }

  private static TemplateScalarModel makeStringAdapter(String aString) {
    return new SimpleScalar(aString);
  }

  private TemplateHashModel makeMapAdapter(Map aMap)  {
    return new MapAdapter(aMap);
  }

  private TemplateListModel makeIteratorAdapter(Iterator anIterator) {
    return new IteratorAdapter(anIterator);
  }

  private TemplateMethodModel makeFunctionAdapter(Generator.Function aFunction) {
    return new FunctionAdapter(aFunction);
  }

  private TemplateHashModel makeBeanAdapter(Object anObject)  {
    return new BeanAdapter(anObject);
  }

  public TemplateModel makeAdapter(Object anObject) throws TemplateModelException {
    Object object = anObject;
    if (interceptor != null) {
      object = interceptor.intercept(object);
    }

    if (object == null) {
      return null;
    }
    if (object instanceof TemplateModel) {
      return (TemplateModel) object;
    }
    else if (object instanceof Generator.Function) {
      return makeFunctionAdapter((Generator.Function) object);
    }
    else if (object instanceof Integer) {
      return makeStringAdapter(object.toString());
    }
    else if (object instanceof Boolean) {
      if (((Boolean) object).booleanValue()) {
        return makeStringAdapter("1");
      }
      else {
        return makeStringAdapter("0");
      }
    }
    else if (object instanceof String) {
      return makeStringAdapter((String) object);
    }
    else if (object instanceof Map) {
      return makeMapAdapter((Map) object);
    }
    else if (object instanceof Iterator) {
      return makeIteratorAdapter((Iterator) object);
    }
    else if (object instanceof Collection) {
      return makeIteratorAdapter(((Collection) object).iterator());
    }
    else if (object instanceof Number) {
      return makeAdapter(new GeneratorFormatAdapters.NumberFormatAdapter((Number) object));
    }
    else if (object instanceof Date) {
      return makeAdapter(new GeneratorFormatAdapters.DateFormatAdapter((Date) object));
    }
    else {
      return makeBeanAdapter(object);
    }
  }

  private class MapAdapter implements TemplateModelRoot {
    private Map map;
    private Map valuesCache;

    private MapAdapter(Map aMap) {
      map = aMap;
      valuesCache = new HashMap();
    }

    public void put(String aKey, TemplateModel aModel) {
      valuesCache.put(aKey, aModel);
    }

    public void remove(String aKey) {
    }

    public boolean isEmpty() {
      return map.isEmpty();
    }

    public TemplateModel get(String aKey) throws TemplateModelException {
      try {
        if (!valuesCache.containsKey(aKey)) {
          Object value = map.get(aKey);

          if (value == null && !map.containsKey(aKey)) {
            throw new TemplateModelException("MapAdapter: no key "+aKey+" available");
          }

          valuesCache.put(aKey, makeAdapter(value));
        }

        return (TemplateModel) valuesCache.get(aKey);
      }
      catch (TemplateModelException e) {
        throw e;
      }
      catch (Throwable t) {
        throw new TemplateModelException(t.getMessage());
      }
    }
  }

  private class IteratorAdapter implements TemplateListModel {
    private Iterator iterator;
    private List valuesCache;
    private int position;

    private IteratorAdapter(Iterator anIterator) {
      iterator = anIterator;

      valuesCache = new ArrayList();
      position=0;


      if (iterator instanceof RewindableIterator) {
        ((RewindableIterator) iterator).rewind();
      }
    }

    public boolean isEmpty() {
      return valuesCache.isEmpty() && !iterator.hasNext();
    }

    private void getUntil(int anIndex) throws TemplateModelException {
      while (valuesCache.size()<=anIndex && iterator.hasNext())
      {
        valuesCache.add(makeAdapter(iterator.next()));
      }
    }

    public TemplateModel get(int anIndex) throws TemplateModelException {
      getUntil(anIndex);

      if (anIndex<valuesCache.size()) {
        return (TemplateModel) valuesCache.get(anIndex);
      }
			throw new TemplateModelException( "Iterator out of bounds" );
    }

    public boolean hasNext() {
      return position<valuesCache.size() || iterator.hasNext();
    }

    public boolean isRewound() {
      return position==0;
    }

    public TemplateModel next() throws TemplateModelException {
      TemplateModel result;

      if (hasNext()) {
        result = get(position);
        position++;
      }
      else
        throw new TemplateModelException( "Iterator out of bounds" );

      return result;
    }

    public void rewind() {
      position=0;
    }
  }

  private class FunctionAdapter implements TemplateMethodModel {
    private Generator.Function function;

    FunctionAdapter(Generator.Function aFunction) {
      function = aFunction;
    }

    public TemplateModel exec(List anArguments) throws TemplateModelException {
      try {
        return makeAdapter(function.perform(anArguments));
      }
      catch (Throwable t) {
        throw new TemplateModelException(t.getMessage());
      }
    }

    public boolean isEmpty() {
      return false;
    }

  }

  private class BeanAdapter implements TemplateHashModel {
    private Object object;

    public BeanAdapter(Object anObject) {
      object = anObject;
    }

    public void put(String aKey, TemplateModel aModel)  throws TemplateModelException  {
      throw new TemplateModelException("FreemarkerGenerator$BeanAdapter.put not supported");
    }

    public void remove(String aKey) throws TemplateModelException  {
      throw new TemplateModelException("FreemarkerGenerator$BeanAdapter.remove not supported");
    }

    public boolean isEmpty() {
      return false;
    }

    public TemplateModel get(String aKey) throws TemplateModelException {
      try {
        if (PropertyUtils.isReadable(object, aKey)) {
          return makeAdapter(PropertyUtils.getSimpleProperty(object, aKey));
        }
				return makeAdapter(MethodUtils.invokeExactMethod(object, "get", aKey));
      }
      catch (Throwable t) {
        throw new TemplateModelException(t.getMessage());
      }
    }
  }

  public static class FreemarkerGeneratorLibrary implements Library {
    private FileTemplateCache templateCache;

    public FreemarkerGeneratorLibrary(File aTemplateRoot) {
      templateCache = new FileTemplateCache(aTemplateRoot);
      templateCache.setLoadingPolicy(FileTemplateCache.LOAD_ON_DEMAND);
    }

    public Generator makeGenerator(String anIdentifier, Interceptor anInterceptor) throws GeneratorExc, GeneratorFailure {
      Template template = (Template) templateCache.getItem(anIdentifier, "template");

      if (template==null) {
        throw new GeneratorExc("FreemarkerGeneratorLibrary: Can't find template " +
            templateCache.getDirectory() + File.separatorChar + anIdentifier);
      }

      return new FreemarkerGenerator(template, anInterceptor);
    }

  }

  public static class FreemarkerGeneratorLibraryFactory implements LibraryFactory {
    private File basePath;

    public FreemarkerGeneratorLibraryFactory(File aBasePath) {
      basePath = aBasePath;
    }

    public Library makeLibrary(String anInitializationString) {
      // todo: the initialization string should be parsed
      return new FreemarkerGeneratorLibrary(new File(basePath, anInitializationString));
    }
  }
}
