/*
 * Copyright (C) 2005 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License.
 * You must obey the GNU General Public License in all respects for all of the code used
 * other than the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mir.generator.tal;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Map;

import mir.generator.tal.interfaces.TALExpressionParser;
import mir.generator.tal.template.CoreTemplateNodeLibrary;
import mir.generator.tal.template.MacroTemplateNodeLibrary;
import mir.generator.tal.template.Template;
import mir.generator.tal.template.TemplateLibrary;
import mir.generator.tal.template.TemplateNode;
import mir.generator.tal.template.TemplateNodeLibrary;
import mir.generator.tal.template.TemplateProcessingException;
import mir.util.StringRoutines;

public class TALTemplateEngine implements TemplateLibrary {
  private TALExpressionParser expressionParser;
  private File basePath;
  private CachingFileLoader loader;

  private SimpleTemplateNodeLibraryRegistry registry;

  public TALTemplateEngine(TALExpressionParser anExpressionParser, File aBasePath) {
    expressionParser = anExpressionParser;
    basePath = aBasePath;
    loader = new CachingFileLoader(100, new TemplateFactory());

    registry = new SimpleTemplateNodeLibraryRegistry();

    registerTemplateNodeLibrary("tal", "http://xml.zope.org/namespaces/tal",
        new CoreTemplateNodeLibrary("tal", "http://xml.zope.org/namespaces/tal"));
    registerTemplateNodeLibrary("metal", "http://xml.zope.org/namespaces/metal",
        new MacroTemplateNodeLibrary("metal", "http://xml.zope.org/namespaces/metal"));
  }

  public void registerTemplateNodeLibrary(String aPrefix, String aUrl, TemplateNodeLibrary aLibrary) {
    registry.registerTemplateNodeLibrary(aPrefix, aUrl, aLibrary);
  }

  public Template loadTemplate(String aName) throws IOException, TemplateProcessingException {
    List parts = StringRoutines.splitString(aName, "#");

    if (parts.size()!=1 && parts.size()!=2) {
      throw new TemplateProcessingException("Invalid template name: " + aName);
    }

    File location = new File(basePath, (String) parts.get(0));

    Template baseTemplate = (Template) loader.retrieveFile(location.getCanonicalPath());

    if (baseTemplate!=null && parts.size()>1) {
      Map definitions = (Map) baseTemplate.getContext().get(MacroTemplateNodeLibrary.MACRO_DEFINITIONS_KEY);
      if (definitions==null || !definitions.containsKey(parts.get(1))) {
        throw new TemplateProcessingException("No macro '"+(String) parts.get(1)+"' found in template '" + (String) parts.get(0) + "'");
      }

      return new Template(expressionParser, (TemplateNode) definitions.get(parts.get(1)), baseTemplate.getContext());
    }
		return baseTemplate;
  }

  public Template lookupTemplate(String aName) throws TemplateProcessingException {
    try {
      return loadTemplate(aName);
    }
    catch (IOException e) {
      throw new TemplateProcessingException("Can't load template " + aName + ": " + e.toString(), e);
    }
  }

  private class TemplateFactory implements CachingFileLoader.CachedFileObjectFactory {
    public Object constructObject(InputStream aStream) {
      try {
        TALTemplateParser parser = new TALTemplateParser(registry);

        return parser.parse(aStream, expressionParser);
      }
      catch (Throwable t) {
        throw new TALFailure(t);
      }
    }
  }
}