/*
 * Copyright (C) 2001, 2002 The Mir-coders group
 *
 * This file is part of Mir.
 *
 * Mir is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Mir is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mir; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * In addition, as a special exception, The Mir-coders gives permission to link
 * the code of this program with  any library licensed under the Apache Software License,
 * The Sun (tm) Java Advanced Imaging library (JAI), The Sun JIMI library
 * (or with modified versions of the above that use the same license as the above),
 * and distribute linked combinations including the two.  You must obey the
 * GNU General Public License in all respects for all of the code used other than
 * the above mentioned libraries.  If you modify this file, you may extend this
 * exception to your version of the file, but you are not obligated to do so.
 * If you do not wish to do so, delete this exception statement from your version.
 */
package mircoders.pdf;

import com.lowagie.text.*;
import com.lowagie.text.pdf.*;
import mir.config.MirPropertiesConfiguration;
import mir.entity.EntityBrowser;
import mir.entity.adapter.EntityAdapter;
import mir.generator.GeneratorExc;
import mir.generator.GeneratorHelper;
import mir.log.LoggerWrapper;
import mir.misc.StringUtil;
import mir.util.DateTimeRoutines;
import mir.util.HTMLRoutines;
import mir.util.ParameterExpander;
import mircoders.entity.EntityContent;
import mircoders.entity.EntityImages;
import mircoders.global.MirGlobal;
import mircoders.localizer.MirLocalizerExc;
import mircoders.storage.DatabaseImages;
import multex.Failure;
import org.apache.oro.text.regex.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.*;
import java.util.List;

public class PDFGenerator {
  public Document document;
  public PdfWriter writer;
  public PdfContentByte cb;
  public String localImageDir;
  public float currentYPosition;
  public int currentPage;
  public float pageWidth;
  public float pageHeight;
  public float verticalMargin;
  public float horizontalMargin;
  public float topEdge;
  public float bottomEdge;
  public float rightEdge;
  public float leftEdge;

  public int maxImageHeight;
  public int maxImageWidth;
  protected LoggerWrapper logger;

  public int indexFontSize;
  public int indexLineHeight;
  public int indexFontFamily;

  public float footerHeight;
  public String footerText;
  public int footerFontSize;
  public int footerFontFamily;

  public int metaHeight;
  public int metaFontSize;
  public int metaFontFamily;

  public int descriptionLineHeight;
  public int descriptionFontSize;
  public int descriptionFontFamily;

  public int contentLineHeight;
  public int contentFontSize;
  public int contentFontFamily;

  public int sourceLineHeight;
  public int sourceFontSize;
  public int sourceFontFamily;

  public int bigImageCaptionFontSize;
  public int bigImageCaptionFontFamily;

  protected MirPropertiesConfiguration configuration;
  private static Pattern nobackslashr;
  private static Pattern HxTag;
  private static Pattern ListItemTag;
  private static Pattern ListTag;
  private static Pattern DivTag;
  private static Pattern PTag;
  private static Pattern PTagClose;
  private static Pattern BRTag;
  private static Pattern ATagAll;
  private static Pattern noTags;
  private static Pattern re1;
  private static Pattern re2;
  private static Pattern re3;

  static {
    Perl5Compiler compiler = new Perl5Compiler();

    try {
      Pattern nobackslashr = compiler.compile("\r");

      HxTag = compiler.compile("</?h[1-6][^>]*>", Perl5Compiler.READ_ONLY_MASK +
                                                     Perl5Compiler.CASE_INSENSITIVE_MASK );
      ListItemTag = compiler.compile("<li[^>]*>", Perl5Compiler.CASE_INSENSITIVE_MASK);
      ListTag = compiler.compile("<(u|o)l[^>]*>", Perl5Compiler.CASE_INSENSITIVE_MASK);
      DivTag = compiler.compile("</?div[^>]*>", Perl5Compiler.CASE_INSENSITIVE_MASK);
      PTag = compiler.compile("<(p|P)([:space:]+[^>]*)?>", Perl5Compiler.READ_ONLY_MASK);
      PTagClose = compiler.compile("</(p|P)([:space:]+[^>]*)?>", Perl5Compiler.READ_ONLY_MASK);
      BRTag = compiler.compile("<(br|BR)([:space:]+[^>]*)?>", Perl5Compiler.READ_ONLY_MASK);
      ATagAll = compiler.compile("<a[^>]*href=(?:\"|\')([^#\"\'][^\'\"]+)(?:\"|\')[^>]*>(.*?)</a>",
                                                     Perl5Compiler.READ_ONLY_MASK +
                                                     Perl5Compiler.CASE_INSENSITIVE_MASK );
      noTags = compiler.compile("<[^>]*>", Perl5Compiler.READ_ONLY_MASK);
      re1 = compiler.compile("\r?\n\r?\n", Perl5Compiler.READ_ONLY_MASK);
      re2 = compiler.compile("\r?\n", Perl5Compiler.READ_ONLY_MASK);
      re3 = compiler.compile("BREAKHERE", Perl5Compiler.READ_ONLY_MASK);
    }
    catch (MalformedPatternException e) {
      throw new RuntimeException(e.getMessage());
    }
  }


  public PDFGenerator(ByteArrayOutputStream out) {
    logger = new LoggerWrapper("PDFGenerator");
    configuration = MirPropertiesConfiguration.instance();
    localImageDir = configuration.getString("Producer.Image.Path");

    try {
      indexFontSize = Integer.parseInt(configuration.getString("PDF.Index.FontSize"));
      indexLineHeight = Integer.parseInt(configuration.getString("PDF.Index.LineHeight"));
      indexFontFamily = getFontByName(configuration.getString("PDF.Index.FontFamily"));

      footerText = configuration.getString("PDF.Footer.String");
      footerFontSize = Integer.parseInt(configuration.getString("PDF.Footer.FontSize"));
      footerFontFamily = getFontByName(configuration.getString("PDF.Footer.FontFamily"));
      footerHeight = Integer.parseInt(configuration.getString("PDF.Footer.Height"));

      metaFontSize = Integer.parseInt(configuration.getString("PDF.Meta.FontSize"));
      metaFontFamily = getFontByName(configuration.getString("PDF.Meta.FontFamily"));
      metaHeight = Integer.parseInt(configuration.getString("PDF.Meta.Height"));

      descriptionFontSize = Integer.parseInt(configuration.getString("PDF.Description.FontSize"));
      descriptionLineHeight = Integer.parseInt(configuration.getString("PDF.Description.LineHeight"));
      descriptionFontFamily = getFontByName(configuration.getString("PDF.Description.FontFamily"));

      contentFontSize = Integer.parseInt(configuration.getString("PDF.Content.FontSize"));
      contentLineHeight = Integer.parseInt(configuration.getString("PDF.Content.LineHeight"));
      contentFontFamily = getFontByName(configuration.getString("PDF.Content.FontFamily"));

      sourceFontSize = Integer.parseInt(configuration.getString("PDF.Source.FontSize"));
      sourceLineHeight = Integer.parseInt(configuration.getString("PDF.Source.LineHeight"));
      sourceFontFamily = getFontByName(configuration.getString("PDF.Source.FontFamily"));

      bigImageCaptionFontSize = Integer.parseInt(configuration.getString("PDF.BigImageCaption.FontSize"));
      bigImageCaptionFontFamily = getFontByName(configuration.getString("PDF.BigImageCaption.FontFamily"));

    }
    catch (NumberFormatException e) {
      throw new Failure(e.getMessage(), e);
    }

    // step 1: make a document

    String pageSize = configuration.getString("PDF.Pagesize");

    if (pageSize.equals("LETTER")) {
      document = new Document(PageSize.LETTER);
      pageWidth = 612;
      pageHeight = 792;
    } else {
      document = new Document(PageSize.A4);
      pageWidth = 595;
      pageHeight = 842;
    }

    maxImageHeight = 250;
    maxImageWidth = 250;


    verticalMargin = 20;
    horizontalMargin = 20;


    topEdge = pageHeight - verticalMargin;
    bottomEdge = verticalMargin;
    rightEdge = pageWidth - horizontalMargin;
    leftEdge = horizontalMargin;

    currentYPosition = topEdge;
    currentPage = 1;

    String headerText = configuration.getString("PDF.Title.String");

    try {
      writer = PdfWriter.getInstance(document, out);
      cb = writer.getDirectContent();

      document.open();
      addHeader(headerText);
    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }
  }

  public void stop() {
    addFooter();
    document.close();
  }

  public void addHeader(String headerText) {
    int titleFontSize = Integer.parseInt(configuration.getString("PDF.Title.FontSize"));
    int titleLineHeight = Integer.parseInt(configuration.getString("PDF.Title.LineHeight"));
    String titleFontFamily = configuration.getString("PDF.Title.FontFamily");
    String headerImage = configuration.getString("PDF.Header.Image");
    int headerImageHeight = Integer.parseInt(configuration.getString("PDF.Header.ImageHeight"));

    try {
      if ((! headerImage.equals("")) && headerImageHeight != 0) {
        PdfTemplate template = cb.createTemplate(rightEdge - horizontalMargin, headerImageHeight);

        float toYPosition = currentYPosition - headerImageHeight;
        Image theImage = Image.getInstance(headerImage);
        theImage.setAbsolutePosition(0, 0);
        //	theImage.scaleAbsolute(img_width,img_height);
        template.addImage(theImage);


        cb.addTemplate(template, leftEdge, toYPosition);
        currentYPosition = toYPosition;
      }
      if (! headerText.equals("")) {
        ColumnText ct = new ColumnText(cb);
        //add a basic header
        ct.addText(new Phrase(headerText, new Font(getFontByName(titleFontFamily), titleFontSize, Font.BOLD)));
        float[] rightCol = {rightEdge, currentYPosition, rightEdge, currentYPosition - titleLineHeight};
        float[] leftCol = {leftEdge, currentYPosition, leftEdge, currentYPosition - titleLineHeight};
        ct.setColumns(leftCol, rightCol);
        ct.setYLine(currentYPosition);
        ct.setAlignment(Element.ALIGN_CENTER);
        ct.go();

        currentYPosition = currentYPosition - titleLineHeight;
      }
    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }
    catch (MalformedURLException de) {
      logger.error(de.getMessage());
    }
    catch (IOException de) {
      logger.error(de.getMessage());
    }
  }

  public void addIndexItem(EntityContent entityContent) {
    try {

      ColumnText ict = new ColumnText(cb);
      String theTitle = entityContent.getFieldValue("title");
      String theCreator = entityContent.getFieldValue("creator");
      Phrase titleP = new Phrase(" - " + theTitle, new Font(indexFontFamily, indexFontSize, Font.BOLD));
      Phrase creatorP = new Phrase(" :: " + theCreator, new Font(indexFontFamily, indexFontSize));
      float toYPosition = currentYPosition - indexLineHeight;
      float[] leftIndexCols = {leftEdge, currentYPosition, leftEdge, toYPosition};
      float[] rightIndexCols = {rightEdge, currentYPosition, rightEdge, toYPosition};
      ict.addText(titleP);
      ict.addText(creatorP);
      ict.setColumns(leftIndexCols, rightIndexCols);
      ict.setYLine(currentYPosition);
      ict.setAlignment(Element.ALIGN_LEFT);
      ict.go();
      currentYPosition = toYPosition;
    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }


  }

  private int addTextLine(ColumnText ct, int lineHeight, int alignment, float left, float right) {
    logger.debug("adding a line of text");
    if (! enoughY(lineHeight)) {
      newPage();
    }
    float toYPosition = currentYPosition - lineHeight;
    float[] leftContentCols = {left, currentYPosition, left, toYPosition};
    float[] rightContentCols = {right, currentYPosition, right, toYPosition};
    ct.setColumns(leftContentCols, rightContentCols);
    ct.setYLine(currentYPosition);
    ct.setAlignment(alignment);
    try {
      int status = ct.go();
      currentYPosition = toYPosition;
      return status;
    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }
    return 0;
  }

  public void addLine() {
    cb.setLineWidth(1f);
    cb.moveTo(rightEdge, currentYPosition - 5);
    cb.lineTo(leftEdge, currentYPosition - 5);
    cb.stroke();
    currentYPosition = currentYPosition - 10;

  }

  public void addFooter() {
    try {
      ColumnText fct = new ColumnText(cb);
      cb.setLineWidth(1f);
      cb.moveTo(rightEdge, bottomEdge + footerHeight - 5);
      cb.lineTo(leftEdge, bottomEdge + footerHeight - 5);
      cb.stroke();
      float[] leftFooterCols = {leftEdge, bottomEdge + footerHeight - 1, leftEdge, bottomEdge};
      float[] rightFooterCols = {rightEdge, bottomEdge + footerHeight - 1, rightEdge, bottomEdge};

      Paragraph footerP = new Paragraph(footerText, new Font(footerFontFamily, footerFontSize));
      fct.addText(footerP);

      fct.setColumns(leftFooterCols, rightFooterCols);
      fct.setYLine(bottomEdge + footerHeight - 1);
      fct.setAlignment(Element.ALIGN_JUSTIFIED);
      fct.go();

      Paragraph numberP = new Paragraph((new Integer(currentPage)).toString(), new Font(footerFontFamily, footerFontSize, Font.BOLD));
      fct.addText(numberP);
      fct.setAlignment(Element.ALIGN_RIGHT);
      fct.go();

    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }


  }

  public void newPage() {
    try {
      //add a footer
      addFooter();
      document.newPage();
      currentPage++;
      currentYPosition = topEdge;
    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }
  }

  public void addArticleSeparator() {
    // make a line
    if (! enoughY(10)) {
      newPage();
    }
    cb.setLineWidth(1f);
    cb.moveTo(rightEdge, currentYPosition - 5);
    cb.lineTo(leftEdge, currentYPosition - 5);
    cb.stroke();
    currentYPosition = currentYPosition - 10;
  }

  public void addArticleMetaInfo(ColumnText ct, String theTitle, String theCreator, String theDate) {
    //see if we have room for the author and title

    if (! enoughY(metaHeight)) {
      newPage();
    }

    Paragraph titleP = new Paragraph(theTitle + "\n", new Font(metaFontFamily, metaFontSize, Font.BOLD));
    Paragraph whowhenP = new Paragraph(theCreator + "  " + theDate, new Font(metaFontFamily, metaFontSize));
    ct.addText(titleP);
    ct.addText(whowhenP);

    ct.setYLine(currentYPosition);
    ct.setAlignment(Element.ALIGN_LEFT);

    float toYPosition = currentYPosition - metaHeight;
    float[] leftHeadCols = {leftEdge, currentYPosition, leftEdge, toYPosition};
    float[] rightHeadCols = {rightEdge, currentYPosition, rightEdge, toYPosition};

    ct.setColumns(leftHeadCols, rightHeadCols);
    try {
      ct.go();
      currentYPosition = toYPosition;
    }
    catch (DocumentException de) {
      logger.error(de.getMessage());
    }

  }

  public void addArticleDescription(ColumnText ct, String theDescription) {
    // Now we add the description, one line at a time, the ct should be empty at this point


    Paragraph descP = new Paragraph(theDescription, new Font(descriptionFontFamily, descriptionFontSize, Font.BOLD));
    ct.addText(descP);

    // every article has a description, so we can assume that:
    int status = ColumnText.NO_MORE_COLUMN;

    int brake = 1000;
    while ((status & ColumnText.NO_MORE_TEXT) == 0 && brake > 0) {
      //there is still text left in the description.
      status = addTextLine(ct, descriptionLineHeight, Element.ALIGN_JUSTIFIED, leftEdge, rightEdge);
      brake--;
    }
    if (brake == 0)
      logger.error("runaway description...try increasing the line height or decreasing the font size");
  }

  public void addArticleContent(ColumnText ct, String theContent, Iterator images) {
    //let's go ahead and add in all the body text
    Paragraph contentP = new Paragraph(theContent, new Font(contentFontFamily, contentFontSize));
    ct.addText(contentP);

    int contentLinesBeforeImages = 3;
    //and assume we have at least one line of text in the content
    int status = ColumnText.NO_MORE_COLUMN;

    // let's add a little bit of text, like a couple of lines
    int x = 0;
    while (((status & ColumnText.NO_MORE_TEXT) == 0) && x < contentLinesBeforeImages) {
      status = addTextLine(ct, contentLineHeight, Element.ALIGN_JUSTIFIED, leftEdge, rightEdge);
      x++;
    }

    boolean addImageOnLeft = true; //we will alternate within articles
    while (images.hasNext()) {

      EntityImages currentImage = (EntityImages) images.next();
      float img_width = (new Float(currentImage.getFieldValue("img_width"))).floatValue();
      float img_height = (new Float(currentImage.getFieldValue("img_height"))).floatValue();
      if (img_height > maxImageHeight) {
        img_width = (new Float((new Float(img_width * (maxImageHeight / img_height))).intValue())).floatValue();
        img_height = maxImageHeight;
      }
      if (img_width > maxImageWidth) {
        img_height = (new Float((new Float(img_height * (maxImageWidth / img_width))).intValue())).floatValue();
        img_width = maxImageWidth;
      }

      String img_title = currentImage.getFieldValue("title");
      String img_path = currentImage.getFieldValue("publish_path");

      if ((status & ColumnText.NO_MORE_TEXT) == 0) {
        // there is still text, so add an image which will have text wrapped around it, then add the text which
        // will be on the left or the right of the image

        float templateMinimumHeight = img_height + 20;
        float templateWidth = img_width + 10;
        float templateHeight = templateMinimumHeight + contentLineHeight - (templateMinimumHeight % contentLineHeight);

        PdfTemplate template = cb.createTemplate(templateWidth, templateHeight);

        //here we need a page check
        if (! enoughY((new Float(templateHeight)).intValue())) {
          //ok, well just fill text to the bottom then
          float toYPosition = bottomEdge + footerHeight;
          float[] leftBottomCols = {leftEdge, currentYPosition, leftEdge, toYPosition};
          float[] rightBottomCols = {rightEdge, currentYPosition, rightEdge, toYPosition};
          ct.setColumns(leftBottomCols, rightBottomCols);
          ct.setYLine(currentYPosition);
          ct.setAlignment(Element.ALIGN_JUSTIFIED);
          try {
            status = ct.go();
          }
          catch (DocumentException de) {
            logger.error(de.getMessage());
          }
          newPage();
        }

        float toYPosition = currentYPosition - templateHeight;

        try {
          Image theImage = Image.getInstance(localImageDir + img_path);
          theImage.scaleAbsolute(img_width, img_height);
          theImage.setAbsolutePosition(5, 13);

          template.addImage(theImage);
          template.beginText();
          BaseFont bf = BaseFont.createFont(BaseFont.HELVETICA, BaseFont.CP1252, BaseFont.NOT_EMBEDDED);
          template.setFontAndSize(bf, 8);
          template.setTextMatrix(5, 3);
          template.showText(img_title);
          template.endText();
        }
        catch (BadElementException de) {
          logger.error(de.getMessage());
        }
        catch (DocumentException de) {
          logger.error(de.getMessage());
        }
        catch (MalformedURLException de) {
          logger.error(de.getMessage());
        }
        catch (IOException de) {
          logger.error(de.getMessage());
        }


        float leftImageTextEdge = leftEdge;
        float rightImageTextEdge = rightEdge;


        if (addImageOnLeft) {
          cb.addTemplate(template, leftEdge, toYPosition);
          leftImageTextEdge = leftEdge + templateWidth + 5;
          addImageOnLeft = false;
        } else {
          cb.addTemplate(template, rightEdge - templateWidth, toYPosition);
          rightImageTextEdge = rightEdge - templateWidth - 5;
          addImageOnLeft = true;
        }

        logger.debug("adding template at " + leftEdge + "," + toYPosition);

        //and fill some text while we are at it

        float[] leftBottomCols = {leftImageTextEdge, currentYPosition, leftImageTextEdge, toYPosition};
        float[] rightBottomCols = {rightImageTextEdge, currentYPosition, rightImageTextEdge, toYPosition};
        ct.setColumns(leftBottomCols, rightBottomCols);
        ct.setYLine(currentYPosition);
        ct.setAlignment(Element.ALIGN_JUSTIFIED);
        try {
          status = ct.go();
          currentYPosition = toYPosition;
        }
        catch (DocumentException de) {
          logger.error(de.getMessage());
        }


      } else {
        //add an image on the left with a big caption to the right
        currentYPosition = currentYPosition - 10;
        float templateMinimumHeight = img_height;
        float templateWidth = img_width;
        float templateHeight = templateMinimumHeight + contentLineHeight - (templateMinimumHeight % contentLineHeight);
        PdfTemplate template = cb.createTemplate(templateWidth, templateHeight);
        if (! enoughY((new Float(templateHeight)).intValue())) {
          newPage();
        }
        float toYPosition = currentYPosition - templateHeight;
        try {
          Image theImage = Image.getInstance(localImageDir + img_path);
          theImage.setAbsolutePosition(0, 13);
          theImage.scaleAbsolute(img_width, img_height);
          template.addImage(theImage);
        }
        catch (BadElementException de) {
          logger.error(de.getMessage());
        }
        catch (DocumentException de) {
          logger.error(de.getMessage());
        }
        catch (MalformedURLException de) {
          logger.error(de.getMessage());
        }
        catch (IOException de) {
          logger.error(de.getMessage());
        }

        cb.addTemplate(template, leftEdge, toYPosition);

        // add a big caption
        ColumnText cct = new ColumnText(cb);
        float[] leftCaptionCols = {leftEdge + templateWidth + 5, currentYPosition - 5, leftEdge + templateWidth + 5, toYPosition};
        float[] rightCaptionCols = {rightEdge, currentYPosition - 5, rightEdge, toYPosition};

        Paragraph captionP = new Paragraph(img_title, new Font(bigImageCaptionFontFamily, bigImageCaptionFontSize, Font.BOLD));
        cct.addText(captionP);
        cct.setColumns(leftCaptionCols, rightCaptionCols);
        cct.setYLine(currentYPosition - 5);
        cct.setAlignment(Element.ALIGN_LEFT);
        try {
          cct.go();
          currentYPosition = toYPosition;
        }
        catch (DocumentException de) {
          logger.error(de.getMessage());
        }
      }
    }

    //add the rest of the text which might be left
    int brake = 10000;
    while ((status & ColumnText.NO_MORE_TEXT) == 0 && brake > 0) {
      status = addTextLine(ct, contentLineHeight, Element.ALIGN_JUSTIFIED, leftEdge, rightEdge);
      brake --;
    }
    if (brake == 0)
      logger.error("runaway content....try decreasing font size or increasing line height");
  }

  private void addArticleSource(ColumnText ct, String theSource) {
    Paragraph sourceP = new Paragraph(theSource, new Font(sourceFontFamily, sourceFontSize, Font.BOLD));
    ct.addText(sourceP);
    addTextLine(ct, sourceLineHeight, Element.ALIGN_RIGHT, leftEdge, rightEdge);
  }


  private boolean enoughY(int heightOfBlockToAdd) {
    return (currentYPosition - heightOfBlockToAdd - footerHeight) >= bottomEdge;
  }


  public void add(EntityContent anArticle, Locale aLocale) throws GeneratorExc, MirLocalizerExc, Exception {
    logger.debug("adding a content Entity");

    /*
     * initialize
     * separate
     * meta info
     * description
     * content
     * --image with text
     * --normal text
     * --image without text
     * source
    */

    List extraTables = new ArrayList();
    extraTables.add("content_x_media cxm");
    Iterator images = new EntityBrowser(
        DatabaseImages.getInstance(), "i", extraTables,
        "cxm.content_id=" + anArticle.getId() + "and cxm.media_id=i.id",
        "i.id desc", 30, -1, 0);

    String isHTML = anArticle.getFieldValue("is_html");
    String theTitle = anArticle.getFieldValue("title");
    String theCreator = anArticle.getFieldValue("creator");
    String theDate = DateTimeRoutines.advancedDateFormat(
        configuration.getString("RDF.Meta.DateFormat"),
        StringUtil.convertMirInternalDateToDate(anArticle.getFieldValue("webdb_create")),
        configuration.getString("Mir.DefaultTimezone"));


    String theDescriptionRaw = anArticle.getFieldValue("description");
    String theContentRaw = anArticle.getFieldValue("content_data");

    MirPropertiesConfiguration configuration = MirPropertiesConfiguration.instance();
    Map variables = GeneratorHelper.makeBasicGenerationData(new Locale[] {
            aLocale ,new Locale(configuration.getString("Mir.Admin.FallbackLanguage", "en"), "")},
            "bundles.open","bundles.open");

    EntityAdapter article =
            MirGlobal.localizer().dataModel().adapterModel().makeEntityAdapter("content", anArticle);
    variables.put("article", article);
    String source = ParameterExpander.expandExpression(
            variables,
            configuration.getString("PDF.Source")
    );
    String theContent = "";
    String theDescription = "";

    if (isHTML.equals("1")) {
        theContent =
              Util.substitute(new Perl5Matcher(), nobackslashr, new Perl5Substitution(""), theContentRaw, Util.SUBSTITUTE_ALL);
        theDescription =
                Util.substitute(new Perl5Matcher(), nobackslashr, new Perl5Substitution(""), theDescriptionRaw, Util.SUBSTITUTE_ALL);

        theContent = Util.substitute(new Perl5Matcher(), HxTag, new Perl5Substitution("\n\n"), theContent, Util.SUBSTITUTE_ALL);
        theDescription = Util.substitute(new Perl5Matcher(), HxTag, new Perl5Substitution("\n\n"), theDescription, Util.SUBSTITUTE_ALL);
        theContent  = Util.substitute(new Perl5Matcher(), ListItemTag, new Perl5Substitution("\n * "), theContent, Util.SUBSTITUTE_ALL);
        theDescription  = Util.substitute(new Perl5Matcher(), ListItemTag, new Perl5Substitution("\n * "), theDescription, Util.SUBSTITUTE_ALL);

        theContent  = Util.substitute(new Perl5Matcher(), ListTag, new Perl5Substitution("\n"), theContent, Util.SUBSTITUTE_ALL);
        theDescription  = Util.substitute(new Perl5Matcher(), ListTag, new Perl5Substitution("\n"), theDescription, Util.SUBSTITUTE_ALL);

        theContent = Util.substitute(new Perl5Matcher(), DivTag, new Perl5Substitution("\n"), theContent, Util.SUBSTITUTE_ALL);
        theDescription = Util.substitute(new Perl5Matcher(), DivTag, new Perl5Substitution("\n"), theDescription, Util.SUBSTITUTE_ALL);

        theContent = Util.substitute(new Perl5Matcher(), PTag, new Perl5Substitution("\n    "), theContent, Util.SUBSTITUTE_ALL);
        theDescription = Util.substitute(new Perl5Matcher(), PTag, new Perl5Substitution("\n    "), theDescription, Util.SUBSTITUTE_ALL);

        theContent = Util.substitute(new Perl5Matcher(), PTagClose, new Perl5Substitution("\n"), theContent, Util.SUBSTITUTE_ALL);
        theDescription = Util.substitute(new Perl5Matcher(), PTagClose, new Perl5Substitution("\n"), theDescription, Util.SUBSTITUTE_ALL);

        theContent = Util.substitute(new Perl5Matcher(), BRTag, new Perl5Substitution("\n"), theContent, Util.SUBSTITUTE_ALL);
        theDescription = Util.substitute(new Perl5Matcher(), BRTag, new Perl5Substitution("\n"), theDescription, Util.SUBSTITUTE_ALL);


        Perl5Matcher matcher = new Perl5Matcher();
        PatternMatcherInput input = new PatternMatcherInput(theContent);
        String theContentCopy = theContent;
        while (matcher.contains(input, ATagAll)) {
          String atagString = matcher.getMatch().group(0);
          String atagStringHref = matcher.getMatch().group(1);
          String atagStringText = matcher.getMatch().group(2);
          int begin = matcher.getMatch().beginOffset(0);
          theContentCopy = 
                  theContentCopy.substring(0, begin) + atagStringText +
                          " [" + atagStringHref + "] " + theContentCopy.substring(begin + atagString.length());
        }
        theContent = theContentCopy;


        input = new PatternMatcherInput(theDescription);

        String theDescriptionCopy = theDescription;
        while (matcher.contains(input, ATagAll)) {
          String atagString = matcher.getMatch().group(0);
          String atagStringHref = matcher.getMatch().group(1);
          String atagStringText = matcher.getMatch().group(2);
          int begin = matcher.getMatch().beginOffset(0);
          theDescriptionCopy = theDescriptionCopy.substring(0, begin) + atagStringText + " [" + atagStringHref + "] " +
                  theDescriptionCopy.substring(begin + atagString.length());
        }
        theDescription = theDescriptionCopy;


        theContent = Util.substitute(new Perl5Matcher(), noTags, new Perl5Substitution(" "), theContent, Util.SUBSTITUTE_ALL);
        theDescription = Util.substitute(new Perl5Matcher(), noTags, new Perl5Substitution(" "), theDescription, Util.SUBSTITUTE_ALL);

        theContent = HTMLRoutines.resolveHTMLEntites(theContent);
        theDescription = HTMLRoutines.resolveHTMLEntites(theDescription);

        String theDescription1 = Util.substitute(new Perl5Matcher(), re1, new Perl5Substitution("BREAKHERE"), theDescription, Util.SUBSTITUTE_ALL);

        String theDescription2 = Util.substitute(new Perl5Matcher(), re2, new Perl5Substitution(" "), theDescription1, Util.SUBSTITUTE_ALL);

        theDescription = Util.substitute(new Perl5Matcher(), re3, new Perl5Substitution("\n    "), theDescription2, Util.SUBSTITUTE_ALL);
    }
    else {
      String theContent1 = Util.substitute(new Perl5Matcher(), re1,
              new Perl5Substitution("BREAKHERE"), theContentRaw, Util.SUBSTITUTE_ALL);
      String theDescription1 = Util.substitute(new Perl5Matcher(), re1,
              new Perl5Substitution("BREAKHERE"), theDescriptionRaw, Util.SUBSTITUTE_ALL);

      String theContent2 = Util.substitute(new Perl5Matcher(), re2,
              new Perl5Substitution(" "), theContent1, Util.SUBSTITUTE_ALL);
      String theDescription2 = Util.substitute(new Perl5Matcher(), re2,
              new Perl5Substitution(" "), theDescription1, Util.SUBSTITUTE_ALL);

      theContent = "    " + Util.substitute(new Perl5Matcher(), re3,
              new Perl5Substitution("\n    "), theContent2, Util.SUBSTITUTE_ALL);
      theDescription = Util.substitute(new Perl5Matcher(), re3,
              new Perl5Substitution("\n    "), theDescription2, Util.SUBSTITUTE_ALL);
    }

    addArticleSeparator();

    ColumnText ct = new ColumnText(cb);

    addArticleMetaInfo(ct, theTitle, theCreator, theDate);
    addArticleDescription(ct, theDescription);
    addArticleContent(ct, theContent, images);
    addArticleSource(ct, source);
  }

  public int getFontByName(String fontName) {
    int theFont = 0;
    if ("helvetica".equalsIgnoreCase(fontName)) {
      theFont = Font.HELVETICA;
    }
    else if ("courier".equalsIgnoreCase(fontName)) {
      theFont = Font.COURIER;
    }
    else if ("times".equalsIgnoreCase(fontName)) {
      theFont = Font.TIMES_ROMAN;
    }
    else {
      logger.error("using helvetica because I can't get font for name: " + fontName);
      theFont = Font.HELVETICA;
    }

    return theFont;
  }
}